@@ -159,7 +159,7 @@ bool write_lockfile(const char *lockfilename, const char *filename, bool modifie
 		return FALSE;
 	}
 
-	lockdata = charalloc(LOCKSIZE);
+	lockdata = nmalloc(LOCKSIZE);
 	memset(lockdata, 0, LOCKSIZE);
 
 	/* This is the lock data we will store (other bytes remain 0x00):
@@ -212,7 +212,7 @@ char *do_lockfile(const char *filename, bool ask_the_user)
 	char *secondcopy = copy_of(filename);
 	size_t locknamesize = strlen(filename) + strlen(locking_prefix) +
 							strlen(locking_suffix) + 3;
-	char *lockfilename = charalloc(locknamesize);
+	char *lockfilename = nmalloc(locknamesize);
 	struct stat fileinfo;
 
 	snprintf(lockfilename, locknamesize, "%s/%s%s%s", dirname(namecopy),
@@ -235,7 +235,7 @@ char *do_lockfile(const char *filename, bool ask_the_user)
 			return NULL;
 		}
 
-		lockbuf = charalloc(LOCKSIZE);
+		lockbuf = nmalloc(LOCKSIZE);
 
 		readamt = read(lockfd, lockbuf, LOCKSIZE);
 
@@ -258,7 +258,7 @@ char *do_lockfile(const char *filename, bool ask_the_user)
 		lockuser[16] = '\0';
 		free(lockbuf);
 
-		pidstring = charalloc(11);
+		pidstring = nmalloc(11);
 		sprintf (pidstring, "%u", (unsigned int)lockpid);
 
 		/* Display newlines in filenames as ^J. */
@@ -273,7 +273,7 @@ char *do_lockfile(const char *filename, bool ask_the_user)
 		else if (room < breadth(filename)) {
 			char *fragment = display_string(filename,
 								breadth(filename) - room + 3, room, FALSE, FALSE);
-			postedname = charalloc(strlen(fragment) + 4);
+			postedname = nmalloc(strlen(fragment) + 4);
 			strcpy(postedname, "...");
 			strcat(postedname, fragment);
 			free(fragment);
@@ -282,7 +282,7 @@ char *do_lockfile(const char *filename, bool ask_the_user)
 
 		/* Allow extra space for username (14), program name (8), PID (8),
 		 * and terminating \0 (1), minus the %s (2) for the file name. */
-		promptstr = charalloc(strlen(question) + 29 + strlen(postedname));
+		promptstr = nmalloc(strlen(question) + 29 + strlen(postedname));
 		sprintf(promptstr, question, postedname, lockuser, lockprog, pidstring);
 		free(postedname);
 		free(pidstring);
@@ -314,7 +314,7 @@ char *do_lockfile(const char *filename, bool ask_the_user)
 void stat_with_alloc(const char *filename, struct stat **pstat)
 {
 	if (*pstat == NULL)
-		*pstat = (struct stat *)nmalloc(sizeof(struct stat));
+		*pstat = nmalloc(sizeof(struct stat));
 
 	if (stat(filename, *pstat) != 0) {
 		free(*pstat);
@@ -615,7 +615,7 @@ void read_file(FILE *f, int fd, const char *filename, bool undoable)
 		/* The length of the current line of the file. */
 	size_t bufsize = LUMPSIZE;
 		/* The size of the line buffer; increased as needed. */
-	char *buf = charalloc(bufsize);
+	char *buf = nmalloc(bufsize);
 		/* The buffer in which we assemble each line of the file. */
 	linestruct *topline;
 		/* The top of the new buffer where we store the read file. */
@@ -689,7 +689,7 @@ void read_file(FILE *f, int fd, const char *filename, bool undoable)
 			 * decreasing it -- it gets freed when reading is finished. */
 			if (len == bufsize) {
 				bufsize += LUMPSIZE;
-				buf = charealloc(buf, bufsize);
+				buf = nrealloc(buf, bufsize);
 			}
 
 			continue;
@@ -719,7 +719,6 @@ void read_file(FILE *f, int fd, const char *filename, bool undoable)
 
 #ifndef NANO_TINY
 	block_sigwinch(FALSE);
-#endif
 
 	/* When reading from stdin, restore the terminal and reenter curses mode. */
 	if (isendwin()) {
@@ -728,6 +727,7 @@ void read_file(FILE *f, int fd, const char *filename, bool undoable)
 		terminal_init();
 		doupdate();
 	}
+#endif
 
 	/* If there was a real error during the reading, let the user know. */
 	if (ferror(f) && errornumber != EINTR && errornumber != 0)
@@ -889,7 +889,7 @@ char *get_next_filename(const char *name, const char *suffix)
 
 	/* Reserve space for: the name plus the suffix plus a dot plus
 	 * possibly five digits plus a null byte. */
-	buf = charalloc(wholenamelen + 7);
+	buf = nmalloc(wholenamelen + 7);
 	sprintf(buf, "%s%s", name, suffix);
 
 	while (TRUE) {
@@ -916,7 +916,7 @@ static pid_t pid_of_command = -1;
 		/* The PID of a forked process -- needed when wanting to abort it. */
 
 /* Send an unconditional kill signal to the running external command. */
-RETSIGTYPE cancel_the_command(int signal)
+void cancel_the_command(int signal)
 {
 	kill(pid_of_command, SIGKILL);
 }
@@ -1307,7 +1307,7 @@ char *get_full_path(const char *origpath)
 	if (origpath == NULL)
 		return NULL;
 
-	allocation = charalloc(PATH_MAX + 1);
+	allocation = nmalloc(PATH_MAX + 1);
 	here = getcwd(allocation, PATH_MAX + 1);
 
 	/* If getting the current directory failed, go up one level and try again,
@@ -1321,7 +1321,7 @@ char *get_full_path(const char *origpath)
 	/* If we found a directory, make sure its path ends in a slash. */
 	if (here != NULL) {
 		if (strcmp(here, "/") != 0) {
-			here = charealloc(here, strlen(here) + 2);
+			here = nrealloc(here, strlen(here) + 2);
 			strcat(here, "/");
 		}
 	} else {
@@ -1340,7 +1340,7 @@ char *get_full_path(const char *origpath)
 		size_t length = strlen(target);
 
 		if (target[length - 1] != '/') {
-			target = charealloc(target, length + 2);
+			target = nrealloc(target, length + 2);
 			strcat(target, "/");
 		}
 	}
@@ -1367,14 +1367,14 @@ char *get_full_path(const char *origpath)
 		} else {
 			free(target);
 
-			allocation = charalloc(PATH_MAX + 1);
+			allocation = nmalloc(PATH_MAX + 1);
 			target = getcwd(allocation, PATH_MAX + 1);
 
 			/* If we got a result, make sure it ends in a slash.
 			 * Otherwise, ensure that we return NULL. */
 			if (target != NULL) {
 				if (strcmp(target, "/") != 0) {
-					target = charealloc(target, strlen(target) + 2);
+					target = nrealloc(target, strlen(target) + 2);
 					strcat(target, "/");
 				}
 			} else {
@@ -1393,7 +1393,7 @@ char *get_full_path(const char *origpath)
 	/* If we were given more than a bare path, concatenate the target path
 	 * with the filename portion to get the full, absolute file path. */
 	if (!path_only && target != NULL) {
-		target = charealloc(target, strlen(target) + strlen(just_filename) + 1);
+		target = nrealloc(target, strlen(target) + strlen(just_filename) + 1);
 		strcat(target, just_filename);
 	}
 
@@ -1439,7 +1439,7 @@ char *safe_tempfile(FILE **stream)
 	if (tempdir == NULL)
 		tempdir = copy_of("/tmp/");
 
-	tempfile_name = charealloc(tempdir, strlen(tempdir) + 12);
+	tempfile_name = nrealloc(tempdir, strlen(tempdir) + 12);
 	strcat(tempfile_name, "nano.XXXXXX");
 
 	fd = mkstemp(tempfile_name);
@@ -1465,7 +1465,7 @@ void init_operating_dir(void)
 		die(_("Invalid operating directory: %s\n"), operating_dir);
 
 	free(operating_dir);
-	operating_dir = charealloc(target, strlen(target) + 1);
+	operating_dir = nrealloc(target, strlen(target) + 1);
 }
 
 /* Check whether the given path is outside of the operating directory.
@@ -1514,7 +1514,7 @@ void init_backup_dir(void)
 		die(_("Invalid backup directory: %s\n"), backup_dir);
 
 	free(backup_dir);
-	backup_dir = charealloc(target, strlen(target) + 1);
+	backup_dir = nrealloc(target, strlen(target) + 1);
 }
 #endif
 
@@ -1556,10 +1556,11 @@ int copy_file(FILE *inn, FILE *out, bool close_out)
 bool make_backup_of(char *realname)
 {
 	FILE *original = NULL, *backup_file = NULL;
-	int creation_flags, descriptor, verdict;
 	static struct timespec filetime[2];
+	int creation_flags, descriptor;
 	bool second_attempt = FALSE;
 	char *backupname = NULL;
+	int verdict = 0;
 
 	/* Remember the original file's access and modification times. */
 	filetime[0].tv_sec = openfile->statinfo->st_atime;
@@ -1571,7 +1572,7 @@ bool make_backup_of(char *realname)
 	 * by appending a tilde to the original file name.  Otherwise,
 	 * we create a numbered backup in the specified directory. */
 	if (backup_dir == NULL) {
-		backupname = charalloc(strlen(realname) + 2);
+		backupname = nmalloc(strlen(realname) + 2);
 		sprintf(backupname, "%s~", realname);
 	} else {
 		char *thename = get_full_path(realname);
@@ -1586,7 +1587,7 @@ bool make_backup_of(char *realname)
 		} else
 			thename = copy_of(tail(realname));
 
-		backupname = charalloc(strlen(backup_dir) + strlen(thename) + 1);
+		backupname = nmalloc(strlen(backup_dir) + strlen(thename) + 1);
 		sprintf(backupname, "%s%s", backup_dir, thename);
 		free(thename);
 
@@ -1678,7 +1679,7 @@ bool make_backup_of(char *realname)
 		warn_and_briefly_pause(_("Trying again in your home directory"));
 		currmenu = MMOST;
 
-		backupname = charalloc(strlen(homedir) + strlen(tail(realname)) + 9);
+		backupname = nmalloc(strlen(homedir) + strlen(tail(realname)) + 9);
 		sprintf(backupname, "%s/%s~XXXXXX", homedir, tail(realname));
 
 		descriptor = mkstemp(backupname);
@@ -2066,11 +2067,9 @@ int do_writeout(bool exiting, bool withprompt)
 		functionptrtype func;
 		const char *msg;
 #ifndef NANO_TINY
-		const char *formatstr, *backupstr;
-
-		formatstr = (openfile->fmt == DOS_FILE) ? _(" [DOS Format]") :
+		const char *formatstr = (openfile->fmt == DOS_FILE) ? _(" [DOS Format]") :
 						(openfile->fmt == MAC_FILE) ? _(" [Mac Format]") : "";
-		backupstr = ISSET(MAKE_BACKUP) ? _(" [Backup]") : "";
+		const char *backupstr = ISSET(MAKE_BACKUP) ? _(" [Backup]") : "";
 
 		/* When the mark is on, offer to write the selection to disk, but
 		 * not when in restricted mode, because it would allow writing to
@@ -2214,7 +2213,7 @@ int do_writeout(bool exiting, bool withprompt)
 					char *question = _("File \"%s\" exists; OVERWRITE? ");
 					char *name = display_string(answer, 0,
 										COLS - breadth(question) + 1, FALSE, FALSE);
-					char *message = charalloc(strlen(question) +
+					char *message = nmalloc(strlen(question) +
 												strlen(name) + 1);
 
 					sprintf(message, question, name);
@@ -2332,7 +2331,7 @@ char *real_dir_from_tilde(const char *path)
 #endif
 	}
 
-	retval = charalloc(strlen(tilded) + strlen(path + i) + 1);
+	retval = nmalloc(strlen(tilded) + strlen(path + i) + 1);
 	sprintf(retval, "%s%s", tilded, path + i);
 
 	free(tilded);
@@ -2356,11 +2355,13 @@ int diralphasort(const void *va, const void *vb)
 	if (!aisdir && bisdir)
 		return 1;
 
-	/* Standard function brain damage: We should be sorting
-	 * alphabetically and case-insensitively according to the current
-	 * locale, but there's no standard strcasecoll() function, so we
-	 * have to use multibyte strcasecmp() instead. */
-	return mbstrcasecmp(a, b);
+	int difference = mbstrcasecmp(a, b);
+
+	/* If two names are equivalent when ignoring case, compare them bytewise. */
+	if (difference == 0)
+		return strcmp(a, b);
+	else
+		return difference;
 }
 #endif
 
@@ -2395,8 +2396,8 @@ char **username_completion(const char *morsel, size_t length, size_t *num_matche
 			if (outside_of_confinement(userdata->pw_dir, TRUE))
 				continue;
 #endif
-			matches = (char **)nrealloc(matches, (*num_matches + 1) * sizeof(char *));
-			matches[*num_matches] = charalloc(strlen(userdata->pw_name) + 2);
+			matches = nrealloc(matches, (*num_matches + 1) * sizeof(char *));
+			matches[*num_matches] = nmalloc(strlen(userdata->pw_name) + 2);
 			sprintf(matches[*num_matches], "~%s", userdata->pw_name);
 			++(*num_matches);
 		}
@@ -2444,7 +2445,7 @@ char **filename_completion(const char *morsel, size_t length, size_t *num_matche
 		dirname = real_dir_from_tilde(dirname);
 		/* A non-absolute path is relative to the current browser directory. */
 		if (dirname[0] != '/') {
-			dirname = charealloc(dirname, strlen(present_path) + strlen(wasdirname) + 1);
+			dirname = nrealloc(dirname, strlen(present_path) + strlen(wasdirname) + 1);
 			sprintf(dirname, "%s%s", present_path, wasdirname);
 		}
 		free(wasdirname);
@@ -2470,7 +2471,7 @@ char **filename_completion(const char *morsel, size_t length, size_t *num_matche
 		if (strncmp(entry->d_name, filename, filenamelen) == 0 &&
 									strcmp(entry->d_name, ".") != 0 &&
 									strcmp(entry->d_name, "..") != 0) {
-			fullname = charealloc(fullname, strlen(dirname) +
+			fullname = nrealloc(fullname, strlen(dirname) +
 											strlen(entry->d_name) + 1);
 
 			sprintf(fullname, "%s%s", dirname, entry->d_name);
@@ -2482,7 +2483,7 @@ char **filename_completion(const char *morsel, size_t length, size_t *num_matche
 			if (currmenu == MGOTODIR && !is_dir(fullname))
 				continue;
 
-			matches = (char **)nrealloc(matches, (*num_matches + 1) * sizeof(char *));
+			matches = nrealloc(matches, (*num_matches + 1) * sizeof(char *));
 			matches[*num_matches] = copy_of(entry->d_name);
 			++(*num_matches);
 		}
@@ -2553,7 +2554,7 @@ char *input_tab(char *morsel, size_t *place, void (*refresh_func)(void), bool *l
 		common_len += len1;
 	}
 
-	shared = charalloc(length_of_path + common_len + 1);
+	shared = nmalloc(length_of_path + common_len + 1);
 
 	strncpy(shared, morsel, length_of_path);
 	strncpy(shared + length_of_path, matches[0], common_len);
@@ -2562,7 +2563,7 @@ char *input_tab(char *morsel, size_t *place, void (*refresh_func)(void), bool *l
 	shared[common_len] = '\0';
 
 	/* Cover also the case of the user specifying a relative path. */
-	glued = charalloc(strlen(present_path) + common_len + 1);
+	glued = nmalloc(strlen(present_path) + common_len + 1);
 	sprintf(glued, "%s%s", present_path, shared);
 
 	if (num_matches == 1 && (is_dir(shared) || is_dir(glued)))
@@ -2570,7 +2571,7 @@ char *input_tab(char *morsel, size_t *place, void (*refresh_func)(void), bool *l
 
 	/* If the matches have something in common, copy that part. */
 	if (common_len != *place) {
-		morsel = charealloc(morsel, common_len + 1);
+		morsel = nrealloc(morsel, common_len + 1);
 		strncpy(morsel, shared, common_len);
 		morsel[common_len] = '\0';
 		*place = common_len;
